/**
 * Main class that showcases each design pattern
 */
public class DesignPatternDemo {
  // This class is not an ADT

  public static void main(String[] args) {
    // Idea: Comment out any calls that you haven't implemented the dependencies
    // for yet.
    builderDemo();
    factoryDemo();
    decoratorDemo();
    singletonDemo();
    interningDemo();
    proxyDemo();
  }

  /**
   * Showcases the string builder by comparing concatenating many strings using
   * the naive concatenation operator method and using the SimpleStringBuilder.
   *
   * When you're trying to manipulate immutable objects, the change is performed
   * on the object by return a new version with the changes. Hence, when you are
   * performing many continuous mutation on an immutable object, the overhead
   * cost for allocating new space for it can increase significantly.
   *
   * An advantage of the builder pattern is that it represents the immutable
   * object in a mutable way (in the StringBuilder case, it's a char array).
   * This means that the builder can reduce the overhead cost of allocating a
   * new object.
   *
   * Think of other situations where a builder could be useful, here is one to
   * get you thinking about it:
   * - An ADT that stores a complicated and large data structure, with a copy
   *   constructor
   * - Any time the data required to construct a complete copy of a data
   *   structure can only be applied incrementally because of its availability.
   */
  private static void builderDemo() {
    header("Builder");

    int limit = 100000;
    long start = 0;

    String noBuilder = "";
    SimpleStringBuilder stringBuilder = new SimpleStringBuilder();

    start = System.currentTimeMillis();
    for (int i = 0; i < limit; i++) {
      noBuilder += "a";
    }
    System.out.println("No builder: " + (System.currentTimeMillis() - start) + " milliseconds");

    start = System.currentTimeMillis();
    for (int i = 0; i < limit; i++) {
      stringBuilder.append("a");
    }
    System.out.println("With builder: " + (System.currentTimeMillis() - start) + " milliseconds");

    assert noBuilder.equals(stringBuilder.toString());
  }

  /**
   * Showcases the ice cream factory by constructing a few ice creams.
   *
   * In this case, the factory pattern is useful in that it abstracts away the
   * need to look at any of subclasses to ice cream. More specifically, it
   * gives the user the same amount of control over what IceCream gets created,
   * without the need to understand the constructors to each IceCream.
   *
   * This is especially useful for when an ADT has a constructor (or
   * constructors) with multiple similarly typed parameters or with a sizeable
   * number of paramters.
   *
   * Can you think of more situations where a factory would be useful?
   *
   * To make this work: Implement IceCreamFactory.
   */
  private static void factoryDemo() {
    header("Factory");

    IceCream myIceCream;

    myIceCream = IceCreamFactory.makeIceCreamWithFlavor("CHOCoLATE");
    if (myIceCream == null) System.out.println("No such flavor");
    else System.out.println(myIceCream);

    myIceCream = IceCreamFactory.makeIceCreamWithWaffleCone("VAnIlLA");
    if (myIceCream == null) System.out.println("No such flavor");
    else System.out.println(myIceCream);

    myIceCream = IceCreamFactory.makeLactoseFreeIceCream("neapolitan");
    if (myIceCream == null) System.out.println("No such flavor");
    else System.out.println(myIceCream);

    myIceCream = IceCreamFactory.makeLactoseFreeIceCreamOnWaffle("neapolitan");
    if (myIceCream == null) System.out.println("No such flavor");
    else System.out.println(myIceCream);


    myIceCream = IceCreamFactory.makeLactoseFreeIceCreamOnWaffle("strawberry");
    if (myIceCream == null) System.out.println("No such flavor");
    else System.out.println(myIceCream);
  }

  /**
   * Showcases the decorator pattern by comparing the "decorated" string builder
   * and the ordinary string builder.
   *
   * The decorator pattern is very straight forward in terms of what it's
   * useful for. In this case, we take our existing StringBuilder and "decorate"
   * it with the ability to append commas in between each new string we append.
   *
   * Another neat trick is in PrintingFileManager, another decorator we didn't
   * show off.  One can chain together any number of PrintingFileManagers
   * together by continually passing them into each other's constructors. You
   * might imagine more useful decorators that could be applied multiple times
   * to the same ADT, for instance applying a border to an image.
   *
   * Can you think of other use cases for a decorator?
   *
   * To make this work: Implement SimpleStringBuilder and CSVStringBuilder.
   */
  private static void decoratorDemo() {
    header("Decorator");

    int limit = 100000;
    long start;

    SimpleStringBuilder stringBuilder = new SimpleStringBuilder();
    CSVStringBuilder csvStringBuilder = new CSVStringBuilder(",");

    start = System.currentTimeMillis();
    for (int i = 0; i < limit; i++) {
      stringBuilder.append("a");
    }
    System.out.println("StringBuilder: " + (System.currentTimeMillis() - start) + " milliseconds");

    start = System.currentTimeMillis();
    for (int i = 0; i < limit; i++) {
      csvStringBuilder.append("a");
    }
    System.out.println("CSVStringBuilder: " + (System.currentTimeMillis() - start) + " milliseconds");

    System.out.println(stringBuilder.toString().substring(0, 11));
    System.out.println(csvStringBuilder.toString().substring(0, 11));
  }

  /**
   * Showcases the singleton pattern by getting the singleton instance twice and
   * using the references to demostrate that they are the same object.
   *
   * A singleton is useful when you only want ONE instance of a specific object
   * ever existing. In this case, there's only one MainFileManager ever being
   * used. This allows us to have multiple references to it, while still being
   * guaranteed that we're working on the same object.
   *
   * A more complex use case example is where we have three different methods,
   * each having its own `MainFileManager.getInstance()`. The first method could
   * read a file, the second method could write the file, and the third one
   * could copy (or save) the most recent file edited. By using a singleton,
   * all three methods will be able to see the same most recently accessed file
   * by the manager without having been passed a reference to the manager that
   * accessed the file.
   *
   * To make this work: Implement SimpleStringBuilder, SimpleFileManager, and
   *                    MainFileManager.
   */
  private static void singletonDemo() {
    header("Singleton");

    MainFileManager managerOne = MainFileManager.getInstance();
    MainFileManager managerTwo = MainFileManager.getInstance();

    String content1 = managerOne.get("files/hello_world.txt");

    System.out.println(content1);
    System.out.println(managerTwo.getMostRecentFile());

    String content2 = managerTwo.get("files/change_me.txt");

    System.out.println(content2);
    System.out.println(managerOne.getMostRecentFile());
  }

  /**
   * Showcases the caching behavior of the interning file manager.
   *
   * The interning pattern here is using a cache for the file manager. This
   * makes it so that when we are accessing a file, it is faster in several
   * ways.
   *
   * Firstly, since we are caching the file contents after the first time we
   * read it, "reading" it again simply returns whatever content is in the
   * cache, avoiding having to read the file again.
   *
   * Secondly, there are two writing actions. The first write is `put`, which
   * simply writes the new content to the data stored in the cache. The second
   * write is `update`, where the contents in the cache is then written to the
   * file on disk. This makes it so while we are working with a file, "writing"
   * and "reading" from it is efficient, since we are not performing any disk
   * reads and write, and the slowness of disk operations will only be done at
   * the very beginning or at the very end.
   *
   * In short, interning makes it so some operations can be performed faster
   * while still displaying similar desired behaviors.
   *
   * Can you come up with another more instances where we can use interning?
   *
   * To make this work: Implement SimpleStringBuilder and SimpleFileManager.
   */
  private static void interningDemo() {
    header("Interning");

    SimpleFileManager fileManager = new SimpleFileManager();
    String file = "files/change_me.txt";

    System.out.println(fileManager.get(file));

    fileManager.put(file, "OxCAFEFOOD");

    System.out.println(fileManager.get(file));
    System.out.println(fileManager.cacheContent(file));

    fileManager.update(file);

    System.out.println(fileManager.get(file));
  }

  /**
   * Showcases the alternate data storing of the encrypting file manager.
   *
   * Proxying when done correctly, would produce an ADT with the same interface
   * and behavior as the ADT that it is proxying. The internal behavior,
   * however, can be different.
   *
   * In our case, the proxy for our file manager also encrypts the internal
   * content of the cache. This is particularly useful when the file managers
   * are being used over networks, where the cache is also stored on the network
   * somewhere. By encrypting whatever is in the cache, anyone who's trying to
   * intercept read's and write's to files over the network won't be able to
   * access the actual content of the file.
   *
   * However, the two file managers on either ends (sending or receiving) can
   * still access the contents as if the were using the simple file manager.
   *
   * Can you come up with more situations where proxying is useful?
   *
   * To make this work: Implement SimpleStringBuilder, SimpleFileManager, and
   *                    EncryptingFileManager.
   */
  private static void proxyDemo() {
    header("Proxy");
    String file = "files/hello_world.txt";

    EncryptingFileManager encryptingFileManager = new EncryptingFileManager();
    SimpleFileManager simpleFileManager = new SimpleFileManager();

    String content1 = encryptingFileManager.get(file);
    String content2 = simpleFileManager.get(file);

    assert content1.equals(content2);

    // changing the contents of the files
    encryptingFileManager.put(file, content1 + "!!");
    simpleFileManager.put(file, content2 + "!!");

    System.out.println(encryptingFileManager.cacheContent(file));
    System.out.println(simpleFileManager.cacheContent(file));

    // updating the file on disk
    encryptingFileManager.update(file);
    simpleFileManager.update(file);
  }

  /**
   * Prints a nice header named demo.
   */
  private static void header(String demo) {
    System.out.println("-------------------------------");
    System.out.println(demo);
    System.out.println("-------------------------------");
  }
}
