import java.util.*;
import java.io.*;

public class SimpleFileManager implements FileManager {
  // THIS IS AN EXAMPLE OF THE INTERNING PATTERN.

  // AF(this) = A file manager with files to change stored in cache in the
  //            entries where value.dirty is true.
  // RI: cache != null && contains no null keys or null FileData or null
  //     contents in the FileData.

  /** The size of a read buffer. */
  private static final int BUFFER_LEN = 1024;
  /** Performs extensive representation invariant checks when true. */
  private static final boolean DEBUG = true;

  /** A cache to the contents of files that have been looked up so far. */
  protected Map<String, FileData> cache;

  /**
   * @spec.effects Constructs a SimpleFileManager with an empty cache.
   */
  public SimpleFileManager() {
    this.cache = new HashMap<>();

    checkRep();
  }

  @Override
  public String get(String file) {
    checkRep();
    assert file != null : "Illegal null argument.";

    // If the file hasn't been seen yet, read it into the cache.
    if (!this.cache.containsKey(file)) {
      FileReader f;
      try {
        f = new FileReader(file);
      } catch (FileNotFoundException e) {
        return null;
      }

      char[] buf = new char[BUFFER_LEN];
      SimpleStringBuilder fileString = new SimpleStringBuilder();

      // Keep reading the file into the buffer and appending the result to the
      // string builder.  If any IOException occurs, just return null.
      while (true) {
        int res;
        try {
          res = f.read(buf, 0, BUFFER_LEN);
        } catch (IOException e) {
          tryClose(f);
          return null;
        }
        if (res == -1) {
          break;
        }

        // append the newly read string buffer onto the string builder
        // hint: res is the number of characters read onto the buffer

      }

      // Cache the result for future calls.  The "dirty" bit is false, which
      // means the file on the file system is up to date with the cache.
      this.cache.put(file, new FileData(fileString.toString(), false));
      tryClose(f);
    }

    try {
      // Return the cached result.
      // TODO: replace null with proper call
      return null;
    } finally {
      checkRep();
    }
  }

  @Override
  public void put(String file, String contents) {
    checkRep();
    assert file != null : "Illegal null argument.";
    assert contents != null : "Illegal null argument.";

    // Cache the new contents with the "dirty" bit set to true, because the new
    // contents are probably not the same as those on the file system.
    // TODO

    checkRep();
  }

  @Override
  public boolean update(String file) {
    checkRep();
    assert file != null : "Illegal null argument.";

    // If there is no cache result for file, the file, if it exists, is
    // definitely updated.  If there were changes to make,
    // it would be in the cache.
    if (!this.cache.containsKey(file)) {
      return true;
    }

    try {
      // If the "dirty" bit is true, there are changes to
      // update on the filesystem.
      if (this.cache.get(file).dirty) {
        FileWriter f;
        try {
          f = new FileWriter(file);
        } catch(IOException e) {
          return false;
        }

        // Write the cached result to the FileWriter.  If there's an
        // IOException, return false because we were unable to update the
        // file with the given changes.
        // TODO

        // Update the "dirty" bit to signify that the change has been updated.
        // TODO
        tryClose(f);
      }

      return true;
    } finally {
      checkRep();
    }
  }

  /**
   * Returns the content of the file currently stored in the cache
   *
   * @param the file to look for
   * @return the cache content relating to the passed file, empty if file not
   * in cache
   */
  public String cacheContent(String file) {
    return this.cache.getOrDefault(file, new FileData("", false)).contents;
  }

  /**
   * Tries to close the given stream, and does nothing in failure.
   */
  private void tryClose(Closeable stream) {
    try {
      stream.close();
    } catch(IOException e) { }
  }

  /**
   * Checks that the representation invariant holds.
   */
  protected void checkRep() {
    assert this.cache != null;
    if (DEBUG) {
      assert !this.cache.containsKey(null);
      assert !this.cache.values().contains(null);
      for (FileData data : this.cache.values()) {
        assert data.contents != null;
      }
    }
  }

  /**
   * Stores the contents of a file either waiting to be written to the file or
   * already written to the file.  Immutable.
   */
  protected class FileData {
    // AF(this) = Data relating to a file, either with this.contents contents,
    //            this.dirty is true and this.contents is the data waiting to be
    //            written to a file.
    // RI: contents != null

    /** The contents or future contents of the file this represents. */
    public final String contents;
    /** False if the contents of this represent the current state of the file
        this represents. */
    public final boolean dirty;

    /**
     * @param contents The contents of a file.
     * @param dirty Dictates whether the contents of the file are dirty or not.
     * @spec.requires contents != null
     * @spec.effects Constructs a FileData relating to contents that have yet to
     *               be written to the file system if dirty is true, else
     *               constructs a FileData relating to contents that represent
     *               the current state of a file in the file system.
     */
    public FileData(String contents, boolean dirty) {
      assert contents != null;

      this.contents = contents;
      this.dirty = dirty;

      checkRep();
    }

    /**
     * Checks that the representation invariant holds.
     */
    private void checkRep() {
      assert this.contents != null;
    }
  }
}
